/*
** Module   : CallbackData.java
** Abstract : Contains the data necessary for a callback procedure/oo-method to be called. 
**
** Copyright (c) 2019-2022, Golden Code Development Corporation.
**
** -#- -I- --Date-- ------------------------------------------Description-------------------------------------
** 001 OM  20190403 Extracted as standalone class from DataSet.
** 002 CA  20220930 Refactored the callback invocation to be performed via a call-site and InvokeConfig, to
**                  allow caching of the resolved target.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
**
** Additional terms under GNU Affero GPL version 3 section 7:
**
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
**
**   0. Attribution Requirement.
**
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
**
**   1. No License To Use Trademarks.
**
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
**
**   2. No Misrepresentation of Affiliation.
**
**     You may not represent yourself as Golden Code Development Corporation or FWD.
**
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
**
**   3. No Misrepresentation of Source or Origin.
**
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;

/**
 * This is a UNION object that groups together the data necessary for calling a CALLBACK.
 */
public final class CallbackData
{
   /** The procedure/method to be called. */
   private final String procName;
   
   /** The handle to the persistent procedure that will be the context for the method call. */
   private final handle procContext;
   
   /** The object context for the method call, when the callback is set to an OO instance. */
   private final object<?> objContext;

   /** The call-site associated with this callback. */
   private InvokeConfig callSite;
   
   /**
    * The constructor for this immutable object. Instantiates the final members.
    * <p>
    * Note: {@code procContext} and {@code objContext} cannot be simultaneously be {@code null}
    * or not {@code null}.
    *
    * @param   procName
    *          The procedure/method to be called.
    * @param   procContext
    *          The handle to the persistent procedure that will be the context for the method
    *          call.
    * @param   objContext
    *          The object context for the method call, when the callback is set to an OO
    *          instance.
    * @param   callSite
    *          The call-site instance.
    */
   public CallbackData(String procName, handle procContext, object<?> objContext, InvokeConfig callSite)
   {
      this.procName = procName;
      this.procContext = procContext;
      this.objContext = objContext;
      this.callSite = callSite;
   }
   
   /**
    * Get the procedure name (the name of the procedure or method name to be invoked).
    * 
    * @return  the procedure name.
    */
   public String getTarget()
   {
      return procName;
   }
   
   /**
    * Get the procedure context (if an internal procedure was configured, this is the handle to
    * its parent external procedure). If an OO method was configured, this returns {@code null}.
    *
    * @return  the procedure context.
    */
   public handle getProcedure()
   {
      return procContext;
   }
   
   /**
    * Get the object context (if OO method was configured, this is the object instance of the
    * class the method belongs to). If an internal procedure configured instead, this returns
    * {@code null}.
    *
    * @return  the object context.
    */
   public object<?> getObject()
   {
      return objContext;
   }
   
   /**
    * Get the {@link #callSite}.
    * 
    * @return   See above.
    */
   public InvokeConfig getCallSite()
   {
      return callSite;
   }
}
